package hexagon

import org.lwjgl.glfw.GLFW._
import org.lwjgl.opengl.GL11
import org.lwjgl.opengl.GL30

import hexagon.renderer.VAO
import hexagon.resource.Loader
import hexagon.resource.Shader
import hexagon.world.storage.World
import org.lwjgl.opengl.GL15
import hexagon.renderer.InstancedRenderer
import hexagon.renderer.VAOBuilder
import hexagon.world.render.WorldRenderer
import hexagon.renderer.NoDepthTest
import hexagon.renderer.Renderer
import hexagon.renderer.VBO
import hexagon.renderer.InstancedRenderer
import hexagon.renderer.NoDepthTest
import hexagon.renderer.InstancedRenderer

trait Scene {
  /** Returns true to consume the event */
  def processKeys(key: Int, scancode: Int, action: Int, mods: Int): Boolean
  def windowResized(w: Int, h: Int): Unit = {}
  def render(): Unit
  def tick(): Unit
  def unload(): Unit
}

abstract class MenuScene extends Scene {
  val guiShader = Shader.getShader("gui")

  def unload(): Unit = {

  }
}

class MainMenuScene extends MenuScene {
  def processKeys(key: Int, scancode: Int, action: Int, mods: Int): Boolean = {
    if (action == GLFW_PRESS) key match {
      case GLFW_KEY_ENTER => Main.pushScene(new GameScene)
      case _              =>
    }
    true
  }

  override def render(): Unit = {

  }

  override def tick(): Unit = {
  }
}

class PauseMenuScene extends MenuScene {
  def processKeys(key: Int, scancode: Int, action: Int, mods: Int): Boolean = {
    if (action == GLFW_PRESS) key match {
      case GLFW_KEY_ESCAPE => Main.popScene
      case _               =>
    }
    true
  }

  override def render(): Unit = {

  }

  override def tick(): Unit = {
  }
}

class GameScene extends Scene {
  // Camera, player, mousepicker, world, etc.

  val blockShader = Shader.getShader("block")
  val skyShader = Shader.getShader("sky")
  val crosshairShader = Shader.getShader("crosshair")

  private val skyVAO = new VAOBuilder(8).addVBO(VBO(4).floats(0, 2).create().fillFloats(0, Seq(-1, -1, 1, -1, -1, 1, 1, 1))).create()
  private val crosshairVAO = new VAOBuilder(8).addVBO(VBO(4).floats(0, 2).create().fillFloats(0, Seq(0, 0.02f, 0, -0.02f, -0.02f, 0, 0.02f, 0))).create()

  private val skyRenderer = new Renderer(skyVAO, GL11.GL_TRIANGLE_STRIP) with NoDepthTest
  private val crosshairRenderer = new Renderer(crosshairVAO, GL11.GL_LINES) with NoDepthTest

  val world = new World
  val worldRenderer = new WorldRenderer(world)

  val camera = new Camera(70f, Main.windowSize.x.toFloat / Main.windowSize.y, 0.02f, 1000f)
  val player = new Player(camera)
  
  private var moveWithMouse = true
  
  setUniforms
  
  private def setUniforms(): Unit = {
    camera.setProjMatrix(blockShader)
    blockShader.setUniform1i("blocksPerRow", 256 / 32)
 		blockShader.setUniform1f("totalSize", world.totalSize)
    skyShader.setUniformMat4("invProjMatr", camera.invProjMatr)
    skyShader.setUniform2f("windowSize", Main.windowSize.x, Main.windowSize.y)
    crosshairShader.setUniform2f("windowSize", Main.windowSize.x, Main.windowSize.y)
  }

  def processKeys(key: Int, scancode: Int, action: Int, mods: Int): Boolean = {
    if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS) Main.pushScene(new PauseMenuScene)
    true
  }

  override def windowResized(width: Int, height: Int): Unit = {
    camera.aspect = width.toFloat / height
    camera.updateProjMatrix
    camera.setProjMatrix(blockShader)

    skyShader.setUniformMat4("invProjMatr", camera.invProjMatr)
    skyShader.setUniform2f("windowSize", width, height)
    crosshairShader.setUniform2f("windowSize", width, height)
  }

  def render(): Unit = {
    skyShader.enable()
    skyRenderer.render()

    // render world etc.
    worldRenderer.render(camera)

    if (moveWithMouse) {
      crosshairShader.enable()
      crosshairRenderer.render()
    }
  }

  def tick(): Unit = {
    player.tick()
    camera.updateViewMatrix
    camera.updateUniforms(blockShader)
    skyShader.setUniformMat4("invViewMatr", camera.invViewMatr)
    skyShader.setUniform3f("sun", 0, 0, -1)
    blockShader.setUniform3f("sun", 0, 0, -1)
    
    world.tick()
  }

  def unload(): Unit = {

  }
}