package hexagon.world.storage



import hexagon.world.coord.BlockRelChunk
import hexagon.world.coord.ChunkRelWorld
import hexagon.world.render.ChunkRenderer
import hexagon.world.coord.BlockRelWorld

class Chunk(val coords: ChunkRelWorld, val world: World) {
  /* block information arrays
   * methods for updating blocks
   */
  
  private var storage: ChunkStorage = new SparseChunkStorage(this)
  for (i <- 0 until 16; j <- 0 until 16) storage.setBlock(new Block(BlockRelChunk(i, j, 0*(i + j * i) % 16).withChunk(coords), 1))
  
  private var needsUpdate = true
  
  val renderer: ChunkRenderer = new ChunkRenderer(this)
  
  def blocks = storage
  
  def getBlock(coords: BlockRelChunk) = storage.getBlock(coords)
  
  def tick(): Unit = {
    if (needsUpdate) {
      needsUpdate = false
      renderer.updateContent()
    }
  }
  
//  def render(): Unit = {
    // TODO: implement system (in Loader) for rendering stuff, like e.g. blocks, with instancing etc. Then: render a block!
    // It might be a good idea to keep a lot i the buffer and then change the buffer in-place when blocks are added/removed
    // It might also be a good idea to have several buffers so that waiting time can be reduced
    // Consider using GL30.glMapBufferRange(target, offset, length, access)
//  }
}

object Block {
  val neighbors = Seq(
      ( 0, 1, 0),
			( 0,-1, 0),
			( 1, 0, 0),
			( 0, 0, 1),
			(-1, 0, 1),
			(-1, 0, 0),
			( 0, 0,-1),
			( 1, 0,-1))
}

class Block(val coord: BlockRelWorld, val blockType: Byte) {
  
}

trait ChunkStorage {
  def blockType(coord: BlockRelChunk): Byte
  def getBlock(coord: BlockRelChunk): Option[Block]
  def setBlock(block: Block): Unit
  def allBlocks: Seq[Block]
}

class DenseChunkStorage(chunk: Chunk) extends ChunkStorage {
  private val blockTypes = new Array[Byte](16*16*16)
  
  def blockType(coord: BlockRelChunk) = blockTypes(coord.value)
  def getBlock(coord: BlockRelChunk) = {
    val _type = blockTypes(coord.value)
    if (_type != 0) Some(new Block(coord.withChunk(chunk.coords), _type)) else None
  }
  def setBlock(block: Block): Unit = blockTypes(block.coord.getBlockRelChunk.value) = block.blockType
  def allBlocks = blockTypes.indices.filter(i => blockTypes(i) != 0).map(i => new Block(BlockRelChunk(i).withChunk(chunk.coords), blockTypes(i)))
}

class SparseChunkStorage(chunk: Chunk) extends ChunkStorage {
  private val blocks = scala.collection.mutable.Map.empty[Short, Block]
  
  def blockType(coord: BlockRelChunk) = blocks.get(coord.value.toShort).map(_.blockType).getOrElse(0)
  def getBlock(coord: BlockRelChunk) = blocks.get(coord.value.toShort)
  def setBlock(block: Block): Unit = blocks(block.coord.getBlockRelChunk.value.toShort) = block
  def allBlocks = blocks.values.toSeq
}