package hexagon.world.storage

import hexagon.world.coord.BlockRelChunk
import hexagon.world.blockType.BlockType

trait ChunkStorage {
  def blockType(coord: BlockRelChunk): BlockType

  def getBlock(coord: BlockRelChunk): Option[Block]
  def setBlock(block: Block): Unit
  def removeBlock(coords: BlockRelChunk): Unit

  def allBlocks: Seq[Block]
  def numBlocks: Int

  def isDense: Boolean
  def toDense(): DenseChunkStorage
  def toSparse(): SparseChunkStorage
}

class DenseChunkStorage(chunk: Chunk) extends ChunkStorage {
  private val blockTypes = new Array[Byte](16 * 16 * 16)
  private var _numBlocks = 0

  def blockType(coord: BlockRelChunk): BlockType = BlockType.byId(blockTypes(coord.value))
  def getBlock(coord: BlockRelChunk): Option[Block] = {
    val _type = blockTypes(coord.value)
    if (_type != 0) Some(new Block(coord.withChunk(chunk.coords), BlockType.byId(_type))) else None
  }
  def setBlock(block: Block): Unit = {
    val idx = block.coord.getBlockRelChunk.value
    if (blockTypes(idx) == 0) _numBlocks += 1
    blockTypes(idx) = block.blockType.id
    if (blockTypes(idx) == 0) _numBlocks -= 1
  }
  def removeBlock(coords: BlockRelChunk): Unit = {
    if (blockTypes(coords.value) != 0) _numBlocks -= 1
    blockTypes(coords.value) = 0
  }
  def allBlocks: Seq[Block] = blockTypes.indices.filter(i => blockTypes(i) != 0).map(i => new Block(BlockRelChunk(i).withChunk(chunk.coords), BlockType.byId(blockTypes(i))))
  def numBlocks: Int = _numBlocks
  def isDense: Boolean = true
  def toDense(): DenseChunkStorage = this
  def toSparse(): SparseChunkStorage = {
    val sparse = new SparseChunkStorage(chunk)
    for (b <- allBlocks) sparse.setBlock(b)
    sparse
  }
}

class SparseChunkStorage(chunk: Chunk) extends ChunkStorage {
  private val blocks = scala.collection.mutable.Map.empty[Short, Block]

  def blockType(coord: BlockRelChunk): BlockType = blocks.get(coord.value.toShort).map(_.blockType).getOrElse(BlockType.Air)
  def getBlock(coord: BlockRelChunk): Option[Block] = blocks.get(coord.value.toShort)
  def setBlock(block: Block): Unit = blocks(block.coord.getBlockRelChunk.value.toShort) = block
  def removeBlock(coords: BlockRelChunk): Unit = blocks -= coords.value.toShort
  def allBlocks: Seq[Block] = blocks.values.toSeq
  def numBlocks: Int = blocks.size
  def isDense: Boolean = false
  def toDense(): DenseChunkStorage = {
    val dense = new DenseChunkStorage(chunk)
    for (b <- blocks) dense.setBlock(b._2)
    dense
  }
  def toSparse(): SparseChunkStorage = this
}
