package hexagon.resource

import scala.collection.Seq

import org.joml.Matrix4f
import org.lwjgl.BufferUtils
import org.lwjgl.opengl.GL20
import org.joml.Vector3f

object Shader {
  private var activeShader: Shader = null
  
  private val shaders = collection.mutable.Map.empty[String, Shader]
  private val matrixBuffer = BufferUtils.createFloatBuffer(16)
  
  def getShader(name: String): Shader = shaders.get(name).getOrElse(null)
  
  def init(): Unit = {
	  cleanUp()
	  Shader("block", shaderParts = Seq("vert", "frag"), attribs = Seq("position", "texCoords", "normal", "blockPos", "blockTex"))
	  Shader("selectedBlock", shaderParts = Seq("vert", "frag"), attribs = Seq("position", "blockPos", "color"))
	  Shader("sky", shaderParts = Seq("vert", "frag"), attribs = Seq("position"))
	  Shader("crosshair", shaderParts = Seq("vert", "frag"), attribs = Seq("position"))
	  Shader("gui", shaderParts = Seq("vert", "frag"), attribs = Seq("position"))
  }
  
  def forEachShader(action: Shader => Unit): Unit = {
		shaders.values.foreach(s => {
			s.enable()
			action(s)
		})
	}
	
	def cleanUp(): Unit = {
		GL20.glUseProgram(0)
		for (s <- shaders.values)
			s.free()
		shaders.clear()
	}
	
	def apply(name: String, shaderParts: Seq[String], attribs: Seq[String] = Seq.empty): Shader = {
	  new Shader(name, shaderParts, attribs)
	}
}

class Shader (val name: String, shaderParts: Seq[String], attribs: Seq[String]) extends Resource {
  private var shaderID: Int = _
	private val uniformLocations = collection.mutable.Map.empty[String, Int]
	private val attributeLocations = collection.mutable.Map.empty[String, Int]
	Shader.shaders += name -> this
	
	reload
	
	def getUniformLocation(name: String): Int = uniformLocations.get(name) match {
	  case Some(loc) => loc
	  case None =>
	    val loc = GL20.glGetUniformLocation(shaderID, name)
	    uniformLocations.put(name, loc)
	    loc
	}
	
	def getAttribLocation(name: String): Int = attributeLocations.get(name) match {
	  case Some(loc) => loc
	  case None =>
		  val loc = GL20.glGetAttribLocation(shaderID, name)
		  attributeLocations.put(name, loc)
	    loc
	}
	
	def bindAttribute(loc: Int, name: String): Unit = {
		GL20.glBindAttribLocation(shaderID, loc, name)
	}
	
	private def setUniform(name: String)(func: Int => Unit) = {
	  enable
		val loc = getUniformLocation(name)
		if (loc != -1) func(loc)
	}
	
	def setUniform1i(name: String, a: Int): Unit = setUniform(name)(GL20.glUniform1i(_, a))
	def setUniform1f(name: String, a: Float): Unit = setUniform(name)(GL20.glUniform1f(_, a))
	def setUniform2f(name: String, a: Float, b: Float): Unit = setUniform(name)(GL20.glUniform2f(_, a, b))
	def setUniform3f(name: String, a: Float, b: Float, c: Float): Unit = setUniform(name)(GL20.glUniform3f(_, a, b, c))
	def setUniform4f(name: String, a: Float, b: Float, c: Float, d: Float): Unit = setUniform(name)(GL20.glUniform4f(_, a, b, c, d))
	
	def setUniformMat4(name: String, matrix: Matrix4f): Unit = {
		setUniform(name)(loc => {
  		matrix.get(Shader.matrixBuffer)
  		Shader.matrixBuffer.rewind()
  		GL20.glUniformMatrix4fv(loc, false, Shader.matrixBuffer)
  		Shader.matrixBuffer.clear()
		})
	}
	
	def reload(): Unit = {
    var b = ShaderBuilder.start(name)
    for (p <- shaderParts) b = b.load(p)
    shaderID = b.bindAttribs(attribs: _*).attatchAll().linkAndFinish()._2
    
    uniformLocations.clear()
    attributeLocations.clear()
  }
	
	def enable(): Unit = {
	  if (Shader.activeShader != this) {
	    Shader.activeShader = this
  		GL20.glUseProgram(shaderID)
	  }
	}
	
	def free(): Unit = {
		GL20.glDeleteProgram(shaderID)
		deleteResource()
	}
}