package hexagon.world.coord

import hexagon.Camera
import hexagon.world.storage.World
import org.joml.Vector3d

sealed abstract class AbstractCoord[T <: AbstractCoord[T]](val x: Double, val y: Double, val z: Double) {
  def +(that: T): T
  def into(vec: Vector3d): Vector3d = vec.set(x, y, z)
  
  override def toString() = f"($x%.3f, $y%.3f, $z%.3f)"
}

class NormalCoord(_x: Double, _y: Double, _z: Double) extends AbstractCoord[NormalCoord](_x, _y, _z) {
  def toCylCoord(reference: NormalCoord, world: World): CylCoord = ???
  def toSkewCylCoord(reference: NormalCoord, world: World): SkewCylCoord = toCylCoord(reference, world).toSkewCylCoord()
  def toBlockCoord(reference: NormalCoord, world: World): BlockCoord = toCylCoord(reference, world).toBlockCoord()
  
  def +(that: NormalCoord) = NormalCoord(x + that.x, y + that.y, z + that.z)
}

class CylCoord(_x: Double, _y: Double, _z: Double) extends AbstractCoord[CylCoord](_x, _y, _z) {
  def toNormalCoord(ref: CylCoord, world: World): NormalCoord = {
    val mult = math.exp((this.y - ref.y) / world.radius)
    val v = (this.z - ref.z) / CoordUtils.y60 * world.halfHexAngle
    val z = math.sin(v)
    val y = math.cos(v)

    val scale = world.radius// / math.sqrt(z * z + y * y)
    NormalCoord((this.x - ref.x) * mult, y * scale * mult - world.radius, z * scale * mult)
  }
  def toSkewCylCoord(): SkewCylCoord = SkewCylCoord(x / CoordUtils.y60, y, z - x * 0.5 / CoordUtils.y60)
  def toBlockCoord(): BlockCoord = toSkewCylCoord().toBlockCoord()
  
  def +(that: CylCoord) = CylCoord(x + that.x, y + that.y, z + that.z)
}

class SkewCylCoord(_x: Double, _y: Double, _z: Double) extends AbstractCoord[SkewCylCoord](_x, _y, _z) {
  def toNormalCoord(reference: CylCoord, world: World): NormalCoord = toCylCoord().toNormalCoord(reference, world)
  def toCylCoord(): CylCoord = CylCoord(x * CoordUtils.y60, y, z + x * 0.5)
  def toBlockCoord(): BlockCoord = BlockCoord(x / CoordUtils.y60, y / 0.5, z / CoordUtils.y60)
  
  def +(that: SkewCylCoord) = SkewCylCoord(x + that.x, y + that.y, z + that.z)
}

class BlockCoord(_x: Double, _y: Double, _z: Double) extends AbstractCoord[BlockCoord](_x, _y, _z) {
  def toNormalCoord(reference: CylCoord, world: World): NormalCoord = toSkewCylCoord().toNormalCoord(reference, world)
  def toCylCoord(): CylCoord = toSkewCylCoord().toCylCoord()
  def toSkewCylCoord(): SkewCylCoord = SkewCylCoord(x * CoordUtils.y60, y * 0.5, z * CoordUtils.y60)
  def toPlaceIntBlockCoord(): PlaceInBlockCoord = PlaceInBlockCoord(x + 0.5 * z, y - 0.5, z + 0.5 * x)
  
  def +(that: BlockCoord) = BlockCoord(x + that.x, y + that.y, z + that.z)
}

class PlaceInBlockCoord(_x: Double, _y: Double, _z: Double) extends AbstractCoord[PlaceInBlockCoord](_x, _y, _z) {
  def toBlockCoord(): BlockCoord = BlockCoord((x - 0.5 * z) * 4 / 3, y + 0.5, (z - 0.5 * x) * 4 / 3)
  
  def +(that: PlaceInBlockCoord) = PlaceInBlockCoord(x + that.x, y + that.y, z + that.z)
}


object NormalCoord {
	def apply(vec: Vector3d) = new NormalCoord(vec.x, vec.y, vec.z)
	def apply(_x: Double, _y: Double, _z: Double) = new NormalCoord(_x, _y, _z)
}

/** NormalCoord with z axis wrapped around a cylinder. The y axis is perp. to the x and z axes and also logarithmic */
object CylCoord {
	def apply(vec: Vector3d) = new CylCoord(vec.x, vec.y, vec.z)
	def apply(_x: Double, _y: Double, _z: Double) = new CylCoord(_x, _y, _z)
}

/** CylCoord with the x axis rotated 30 deg */
object SkewCylCoord {
	def apply(vec: Vector3d) = new SkewCylCoord(vec.x, vec.y, vec.z)
	def apply(_x: Double, _y: Double, _z: Double) = new SkewCylCoord(_x, _y, _z)
}

/** SkewCylCoord with different axis scale */
object BlockCoord {
	def apply(vec: Vector3d) = new BlockCoord(vec.x, vec.y, vec.z)
	def apply(_x: Double, _y: Double, _z: Double) = new BlockCoord(_x, _y, _z)
}

/** BlockCoord with x and z replaced with how far along to x and z axes the vector reaches. y = 0 in the middle of the block. */
object PlaceInBlockCoord {
	def apply(vec: Vector3d) = new PlaceInBlockCoord(vec.x, vec.y, vec.z)
	def apply(_x: Double, _y: Double, _z: Double) = new PlaceInBlockCoord(_x, _y, _z)
}
