package hexagon

import org.lwjgl.glfw.GLFW.GLFW_KEY_ENTER
import org.lwjgl.glfw.GLFW.GLFW_KEY_ESCAPE
import org.lwjgl.glfw.GLFW.GLFW_PRESS
import org.lwjgl.opengl.GL11
import org.lwjgl.glfw.GLFW._

import hexagon.renderer.NoDepthTest
import hexagon.renderer.Renderer
import hexagon.renderer.VAOBuilder
import hexagon.renderer.VBO
import hexagon.resource.Shader
import hexagon.world.render.WorldRenderer
import hexagon.world.storage.World
import hexagon.world.coord.CoordUtils
import org.joml.Vector3i
import hexagon.world.coord.RayTracer
import org.joml.Vector2f
import org.joml.Vector2d
import org.lwjgl.glfw.GLFW
import hexagon.world.storage.Block

trait Scene {
  /** Returns true to consume the event */
  def processKeys(key: Int, scancode: Int, action: Int, mods: Int): Boolean
  def processMouseButtons(button: Int, action: Int, mods: Int): Boolean
  def windowResized(w: Int, h: Int): Unit = {}
  def render(): Unit
  def tick(): Unit
  def unload(): Unit
}

abstract class MenuScene extends Scene {
  val guiShader = Shader.getShader("gui")

  def unload(): Unit = {

  }
}

class MainMenuScene extends MenuScene {
  def processKeys(key: Int, scancode: Int, action: Int, mods: Int): Boolean = {
    if (action == GLFW_PRESS) key match {
      case GLFW_KEY_ENTER => Main.pushScene(new GameScene)
      case _              =>
    }
    true
  }
  
  def processMouseButtons(button: Int, action: Int, mods: Int): Boolean = {
    false
  }

  override def render(): Unit = {

  }

  override def tick(): Unit = {
  }
}

class PauseMenuScene extends MenuScene {
  def processKeys(key: Int, scancode: Int, action: Int, mods: Int): Boolean = {
    if (action == GLFW_PRESS) key match {
      case GLFW_KEY_ESCAPE => Main.popScene
      case _               =>
    }
    true
  }
  
  def processMouseButtons(button: Int, action: Int, mods: Int): Boolean = {
    false
  }

  override def render(): Unit = {

  }

  override def tick(): Unit = {
  }
}

class GameScene extends Scene {
  // Camera, player, mousepicker, world, etc.

	val blockShader = Shader.getShader("block")
	val blockSideShader = Shader.getShader("blockSide")
  val selectedBlockShader = Shader.getShader("selectedBlock")
  val skyShader = Shader.getShader("sky")
  val crosshairShader = Shader.getShader("crosshair")

  private val crosshairVAO = new VAOBuilder(8).addVBO(VBO(4).floats(0, 2).create().fillFloats(0, Seq(0, 0.02f, 0, -0.02f, -0.02f, 0, 0.02f, 0))).create()
  private val crosshairRenderer = new Renderer(crosshairVAO, GL11.GL_LINES) with NoDepthTest

  val world = new World
  val worldRenderer = new WorldRenderer(world)

  val camera = new Camera(70f, Main.windowSize.x.toFloat / Main.windowSize.y, 0.02f, 1000f, world)
	val mousePicker = new RayTracer(world, camera, 7)
  val player = new Player(camera)
	
	private var leftMouseButtonDown = false
	private var rightMouseButtonDown = false
	private var leftMouseButtonCountdown = 0
	private var rightMouseButtonCountdown = 0
	
  setUniforms
  
  private def setUniforms(): Unit = {
	  camera.setProjMatrix(blockShader)
	  camera.setProjMatrix(blockSideShader)
    camera.setProjMatrix(selectedBlockShader)
    blockSideShader.setUniform1i("blocksPerRow", 256 / 32)
    blockShader.setUniform1f("totalSize", World.totalSize)
    blockSideShader.setUniform1f("totalSize", World.totalSize)
 		selectedBlockShader.setUniform1f("totalSize", World.totalSize)
    skyShader.setUniformMat4("invProjMatr", camera.invProjMatr)
    skyShader.setUniform2f("windowSize", Main.windowSize.x, Main.windowSize.y)
    crosshairShader.setUniform2f("windowSize", Main.windowSize.x, Main.windowSize.y)
  }

  def processKeys(key: Int, scancode: Int, action: Int, mods: Int): Boolean = {
    /*if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS) Main.pushScene(new PauseMenuScene)
    else */if (key == GLFW_KEY_M && action == GLFW_PRESS) {
			player.moveWithMouse = !player.moveWithMouse;
			if (player.moveWithMouse) glfwSetInputMode(Main.window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
			else glfwSetInputMode(Main.window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
    }
    true
  }
  
  def processMouseButtons(button: Int, action: Int, mods: Int): Boolean = {
    if (button == 0) {
	    rightMouseButtonDown = action != GLFW.GLFW_RELEASE
    } else if (button == 1) {
      leftMouseButtonDown = action != GLFW.GLFW_RELEASE
    }
    true
  }

  override def windowResized(width: Int, height: Int): Unit = {
    camera.aspect = width.toFloat / height
    camera.updateProjMatrix
    camera.setProjMatrix(blockShader)
    camera.setProjMatrix(blockSideShader)
    camera.setProjMatrix(selectedBlockShader)

    skyShader.setUniformMat4("invProjMatr", camera.invProjMatr)
    skyShader.setUniform2f("windowSize", width, height)
    crosshairShader.setUniform2f("windowSize", width, height)
  }

  def render(): Unit = {
    // render world etc.
    worldRenderer.render(camera)

    if (player.moveWithMouse) {
      crosshairShader.enable()
      crosshairRenderer.render()
    }
  }

  def tick(): Unit = {
    player.tick()
    camera.updateCoords
    camera.updateViewMatrix
    camera.updateUniforms(blockShader)
    camera.updateUniforms(blockSideShader)
    camera.updateUniforms(selectedBlockShader)
    skyShader.setUniformMat4("invViewMatr", camera.invViewMatr)
    skyShader.setUniform3f("sun", 0, 1, -1)
    blockShader.setUniform3f("sun", 0, 1, -1)
    blockSideShader.setUniform3f("sun", 0, 1, -1)
    
    // MousePicker
    mousePicker.setRayFromScreen(if (!player.moveWithMouse) new Vector2f((Main.mousePos.x / Main.windowSize.x * 2 - 1).toFloat, (Main.mousePos.y / Main.windowSize.y * 2 - 1).toFloat) else new Vector2f(0, 0))
    worldRenderer.setSelectedBlockAndSide(mousePicker.trace(c => world.getBlock(c).isDefined))
    
    if (leftMouseButtonCountdown == 0) {
      if (leftMouseButtonDown) {
    	  leftMouseButtonCountdown = 10
        worldRenderer.getSelectedBlockAndSide match {
          case Some((coords1, Some(side))) =>
            val offset = Block.neighborOffsets(side)
            val coords = coords1.offset(offset._1, offset._2, offset._3)
            if (world.getBlock(coords).isEmpty) world.setBlock(new Block(coords, 1))
          case _ =>
        }
      }
    } else leftMouseButtonCountdown -= 1
    
    if (rightMouseButtonCountdown == 0) {
      if (rightMouseButtonDown) {
    	  rightMouseButtonCountdown = 10
        worldRenderer.getSelectedBlockAndSide match {
          case Some((coords, side)) =>
            if (world.getBlock(coords).isDefined) world.removeBlock(coords)
          case None =>
        }
      }
    } else rightMouseButtonCountdown -= 1
    
    world.tick(camera)
  }

  def unload(): Unit = {

  }
}