package hexagon.world.storage

import hexagon.world.coord.BlockRelChunk
import hexagon.world.coord.ChunkRelWorld
import hexagon.world.render.ChunkRenderer
import scala.concurrent.Future

object Chunk {
  val neighborOffsets = Seq(
      ( 0, 1, 0),
      ( 1, 0, 0),
      ( 0, 0, 1),
      (-1, 0, 1),
      ( 0,-1, 0),
      (-1, 0, 0),
      ( 0, 0,-1),
      ( 1, 0,-1))
}

class Chunk(val coords: ChunkRelWorld, val world: World) {
  /* block information arrays
   * methods for updating blocks
   */
  
  val neighbors: Array[Chunk] = Array.tabulate(8)(i => {
    val (dx, dy, dz) = Chunk.neighborOffsets(i)
    val c2 = ChunkRelWorld(coords.X + dx, coords.Y + dy, coords.Z + dz)
    world.getChunk(c2) match {
      case Some(chunk) =>
        chunk.neighbors((i + 4) % 8) = this
        chunk.needsRenderingUpdate = true
        chunk
      case None => null
    }
  })
  
  private var storage: ChunkStorage = new DenseChunkStorage(this)
  
  if (coords.Y == -1) {
    for (i <- 0 until 16; j <- 0 until 16; k <- 0 until 16) if (math.random < 0.6) storage.setBlock(new Block(BlockRelChunk(i, j, k).withChunk(coords), 1))
  } else if (coords.Y < 0) {
    for (i <- 0 until 16; j <- 0 until 16; k <- 0 until 16) storage.setBlock(new Block(BlockRelChunk(i, j, k).withChunk(coords), 1))
  }
  
  private var needsRenderingUpdate = true
  private var _renderer: ChunkRenderer = new ChunkRenderer(this)
  def renderer = _renderer
  
  def blocks = storage
  
  def getBlock(coords: BlockRelChunk) = storage.getBlock(coords)
  
  def setBlock(block: Block): Boolean = {
    val blockCoord = block.coord.getBlockRelChunk
    val before = getBlock(blockCoord)
    storage.setBlock(block)
    if (before == None || before.get != block) {
      modifyBlock(blockCoord)
    }
    true
  }
  
  def removeBlock(coords: BlockRelChunk): Boolean = {
    storage.removeBlock(coords)
    modifyBlock(coords)
    true
  }
  
  private def modifyBlock(coords: BlockRelChunk): Unit = {
    needsRenderingUpdate = true
		val xx = if (coords.cx == 0) -1 else if (coords.cx == 15) 1 else 0
		val yy = if (coords.cy == 0) -1 else if (coords.cy == 15) 1 else 0
		val zz = if (coords.cz == 0) -1 else if (coords.cz == 15) 1 else 0
		
    for (i <- 0 until 8) {
      val off = Chunk.neighborOffsets(i)
      if ((off._1 * xx == 1 || off._2 * yy == 1 || off._3 * zz == 1) && neighbors(i) != null) neighbors(i).needsRenderingUpdate = true
    }
  }
  
  def tick(): Unit = {
    if (needsRenderingUpdate) {
      needsRenderingUpdate = false
      if (renderer == null) {
        if (!isEmpty) _renderer = new ChunkRenderer(this)
      } else if (isEmpty) {
        renderer.unload()
        _renderer = null
      }
      if (renderer != null) renderer.updateContent()
    }
    
    if (storage.isDense) {
      if (storage.numBlocks < 48) {
        storage = storage.toSparse()
      }
    } else {
    	if (storage.numBlocks > 64) {
    		storage = storage.toDense()
    	}
    }
  }
  
  def isEmpty(): Boolean = {
    storage.numBlocks == 0
  }
  
  def unload(): Unit = {
    neighbors.foreach(c => if (c != null) c.needsRenderingUpdate = true)
    if (renderer != null) renderer.unload()
    // and other stuff
  }
  
//  def render(): Unit = {
    // TODO: implement system (in Loader) for rendering stuff, like e.g. blocks, with instancing etc. Then: render a block!
    // It might be a good idea to keep a lot i the buffer and then change the buffer in-place when blocks are added/removed
    // It might also be a good idea to have several buffers so that waiting time can be reduced
    // Consider using GL30.glMapBufferRange(target, offset, length, access)
//  }
}

trait ChunkStorage {
  def blockType(coord: BlockRelChunk): Byte
  
  def getBlock(coord: BlockRelChunk): Option[Block]
  def setBlock(block: Block): Unit
  def removeBlock(coords: BlockRelChunk): Unit
  
  def allBlocks: Seq[Block]
  def numBlocks: Int
  
  def isDense: Boolean
  def toDense(): DenseChunkStorage
  def toSparse(): SparseChunkStorage
}

class DenseChunkStorage(chunk: Chunk) extends ChunkStorage {
  private val blockTypes = new Array[Byte](16*16*16)
  private var _numBlocks = 0
  
  def blockType(coord: BlockRelChunk) = blockTypes(coord.value)
  def getBlock(coord: BlockRelChunk) = {
    val _type = blockTypes(coord.value)
    if (_type != 0) Some(new Block(coord.withChunk(chunk.coords), _type)) else None
  }
  def setBlock(block: Block): Unit = {
    val idx = block.coord.getBlockRelChunk.value
    if (blockTypes(idx) == 0) _numBlocks += 1
    blockTypes(idx) = block.blockType
    if (blockTypes(idx) == 0) _numBlocks -= 1
  }
  def removeBlock(coords: BlockRelChunk): Unit = {
    if (blockTypes(coords.value) != 0) _numBlocks -= 1
    blockTypes(coords.value) = 0
  }
  def allBlocks = blockTypes.indices.filter(i => blockTypes(i) != 0).map(i => new Block(BlockRelChunk(i).withChunk(chunk.coords), blockTypes(i)))
  def numBlocks = _numBlocks
  def isDense = true
  def toDense() = this
  def toSparse() = {
    val sparse = new SparseChunkStorage(chunk)
    for (b <- allBlocks) sparse.setBlock(b)
    sparse
  }
}

class SparseChunkStorage(chunk: Chunk) extends ChunkStorage {
  private val blocks = scala.collection.mutable.Map.empty[Short, Block]
  
  def blockType(coord: BlockRelChunk) = blocks.get(coord.value.toShort).map(_.blockType).getOrElse(0)
  def getBlock(coord: BlockRelChunk) = blocks.get(coord.value.toShort)
  def setBlock(block: Block): Unit = blocks(block.coord.getBlockRelChunk.value.toShort) = block
  def removeBlock(coords: BlockRelChunk): Unit = blocks -= coords.value.toShort
  def allBlocks = blocks.values.toSeq
  def numBlocks = blocks.size
  def isDense = false
  def toDense() = {
    val dense = new DenseChunkStorage(chunk)
    for (b <- blocks) dense.setBlock(b._2)
    dense
  }
  def toSparse() = this
}