package hexagon.world.storage

import hexagon.world.coord.BlockRelChunk
import hexagon.world.coord.ChunkRelWorld
import hexagon.world.render.ChunkRenderer
import scala.concurrent.Future
import hexagon.world.coord.BlockCoord
import hexagon.world.coord.CylCoord

object Chunk {
  val neighborOffsets = Seq(
    (0, 1, 0),
    (1, 0, 0),
    (0, 0, 1),
    (-1, 0, 1),
    (0, -1, 0),
    (-1, 0, 0),
    (0, 0, -1),
    (1, 0, -1))
}

class Chunk(val coords: ChunkRelWorld, val world: World) {
  /* block information arrays
   * methods for updating blocks
   */

  val neighbors: Array[Option[Chunk]] = Array.tabulate(8)(i => {
    val (dx, dy, dz) = Chunk.neighborOffsets(i)
    val c2 = ChunkRelWorld(coords.X + dx, coords.Y + dy, coords.Z + dz)
    world.getChunk(c2) match {
      case Some(chunk) =>
        chunk.neighbors((i + 4) % 8) = Some(this)
        chunk.needsRenderingUpdate = true
        Some(chunk)
      case None => None
    }
  })

  private var storage: ChunkStorage = new DenseChunkStorage(this)

  {
    val column = world.getColumn(coords.getColumnRelWorld).get
    
    val noiseSamples = for (i <- 0 to 4; j <- 0 to 4; k <- 0 to 4) yield {
      val c = BlockCoord(coords.X * 16 + i * 4, coords.Y * 16 + j * 4, coords.Z * 16 + k * 4).toCylCoord()
      world.blockGenerator.genNoiseFromCyl(c) + world.blockDensityGenerator.genNoiseFromCyl(c) * 0.25
    }
    
    def lerp(t: Double, v1: Double, v2: Double) = v1 + t * (v2 - v1)
    
    def superInterp(ii: Int, ij: Int, ik: Int, fi: Double, fj: Double, fk: Double): Double = {
      lerp(fi, lerp(fj, lerp(fk, noiseSamples( ii      * 25 +  ij      * 5 + ik    ),
                                 noiseSamples( ii      * 25 +  ij      * 5 + ik + 1)),
                        lerp(fk, noiseSamples( ii      * 25 + (ij + 1) * 5 + ik    ),
                                 noiseSamples( ii      * 25 + (ij + 1) * 5 + ik + 1))),
               lerp(fj, lerp(fk, noiseSamples((ii + 1) * 25 +  ij      * 5 + ik    ),
                                 noiseSamples((ii + 1) * 25 +  ij      * 5 + ik + 1)),
                        lerp(fk, noiseSamples((ii + 1) * 25 + (ij + 1) * 5 + ik    ),
                                 noiseSamples((ii + 1) * 25 + (ij + 1) * 5 + ik + 1))))
    }

    for (i <- 0 until 16; j <- 0 until 16; k <- 0 until 16) {
      val noise = superInterp(i/4, j/4, k/4, (i%4)/4d, (j%4)/4d, (k%4)/4d)
      if (noise > -0.25) storage.setBlock(new Block(BlockRelChunk(i, j, k).withChunk(coords), 1))
    }
//      if (coords.Y * 16 +  j <= column.heightMap(i)(k)) storage.setBlock(new Block(BlockRelChunk(i, j, k).withChunk(coords), 1))
  }

  private var needsRenderingUpdate = true
  private var _renderer: Option[ChunkRenderer] = Some(new ChunkRenderer(this))
  def renderer: Option[ChunkRenderer] = _renderer

  def blocks: ChunkStorage = storage

  def getBlock(coords: BlockRelChunk): Option[Block] = storage.getBlock(coords)

  def setBlock(block: Block): Boolean = {
    val blockCoord = block.coord.getBlockRelChunk
    val before = getBlock(blockCoord)
    storage.setBlock(block)
    if (before == None || before.get != block) {
      onBlockModified(blockCoord)
    }
    true
  }

  def removeBlock(coords: BlockRelChunk): Boolean = {
    storage.removeBlock(coords)
    onBlockModified(coords)
    true
  }

  private def onBlockModified(coords: BlockRelChunk): Unit = {
    needsRenderingUpdate = true
    val xx = if (coords.cx == 0) -1 else if (coords.cx == 15) 1 else 0
    val yy = if (coords.cy == 0) -1 else if (coords.cy == 15) 1 else 0
    val zz = if (coords.cz == 0) -1 else if (coords.cz == 15) 1 else 0

    for (i <- 0 until 8) {
      val off = Chunk.neighborOffsets(i)
      if (off._1 * xx == 1 || off._2 * yy == 1 || off._3 * zz == 1) neighbors(i).foreach(_.needsRenderingUpdate = true)
    }
  }

  def tick(): Unit = {
    if (needsRenderingUpdate) {
      needsRenderingUpdate = false
      renderer match {
        case Some(r) =>
          if (isEmpty) {
            r.unload()
            _renderer = None
          }
        case None =>
          if (!isEmpty) _renderer = Some(new ChunkRenderer(this))
      }

      renderer.foreach(_.updateContent)
    }

    if (storage.isDense) {
      if (storage.numBlocks < 48) {
        storage = storage.toSparse()
      }
    } else {
      if (storage.numBlocks > 64) {
        storage = storage.toDense()
      }
    }
  }

  def isEmpty(): Boolean = {
    storage.numBlocks == 0
  }

  def unload(): Unit = {
    neighbors.foreach(c => c.foreach(_.needsRenderingUpdate = true))
    renderer.foreach(_.unload)
    // and other stuff
  }

  //  def render(): Unit = {
  // TODO: implement system (in Loader) for rendering stuff, like e.g. blocks, with instancing etc. Then: render a block!
  // It might be a good idea to keep a lot i the buffer and then change the buffer in-place when blocks are added/removed
  // It might also be a good idea to have several buffers so that waiting time can be reduced
  // Consider using GL30.glMapBufferRange(target, offset, length, access)
  //  }
}
