package hexagon.world.storage

import java.util.concurrent.Executors

import scala.collection.mutable.PriorityQueue
import scala.collection.mutable.{ Set => MutableSet }
import scala.concurrent.ExecutionContext

import org.joml.Vector2d

import hexagon.Camera
import hexagon.world.coord.BlockRelWorld
import hexagon.world.coord.ChunkRelWorld
import hexagon.world.coord.ColumnRelWorld
import hexagon.world.coord.CoordUtils
import hexagon.world.coord.CylCoord
import hexagon.world.coord.BlockCoord
import java.util.Random
import hexagon.world.gen.noise.NoiseGenerator

object World {
  /** Max-value: 20 */
  val worldSize = 5
  val ringSize = 1 << worldSize
  val ringSizeMask = ringSize - 1
  val totalSize = 16 * ringSize
  val totalSizeMask = totalSize - 1
  val hexAngle = (2 * math.Pi) / totalSize
  val radius = CoordUtils.y60 / hexAngle
  val circumference = totalSize * CoordUtils.y60
  val chunksLoadedPerTick = 1
  val ticksBetweenColumnLoading = 5
}

class World {
  /* blockStorage
   * other world contents
   * methods for world interaction
   * method for loading/unloading blockColumns
   * block methods that will be forwarded to the correct chunk
   * worldGen references and methods
   * might rename this class to WorldStorage, and make another class World
   */

  var renderDistance: Double = 8 * CoordUtils.y60

  val columns = scala.collection.mutable.Map.empty[Long, ChunkColumn]
  val columnsAtEdge = MutableSet.empty[ColumnRelWorld]
  
  private val random = new Random()
  private[storage] val blockGenerator = new NoiseGenerator(random, 8, 0.2)
  private[storage] val blockDensityGenerator = new NoiseGenerator(random, 8, 0.02)
  private[storage] val biomeHeightMapGenerator = new NoiseGenerator(random, 4, 0.002)

  private var loadColumnsCountdown = 0

  private var chunkLoadingOrigin: CylCoord = null
  private val chunksToLoad = PriorityQueue.empty[(Double, ChunkRelWorld)](Ordering.by(-_._1))
  val chunkLoader = ExecutionContext.fromExecutor(Executors.newFixedThreadPool(10))

  def addChunkToLoadingQueue(coords: ChunkRelWorld): Unit = {
    chunksToLoad.enqueue(makeChunkToLoadTuple(coords))
  }

  private def makeChunkToLoadTuple(coords: ChunkRelWorld) = {
    (chunkLoadingOrigin.distanceSq(BlockCoord(coords.X * 16 + 8, coords.Y * 16 + 8, coords.Z * 16 + 8).toCylCoord), coords)
  }

  def getColumn(coords: ColumnRelWorld): Option[ChunkColumn] = columns.get(coords.value)
  def getChunk(coords: ChunkRelWorld): Option[Chunk] = getColumn(coords.getColumnRelWorld).flatMap(_.getChunk(coords.getChunkRelColumn))
  def getBlock(coords: BlockRelWorld): Option[Block] = getColumn(coords.getColumnRelWorld).flatMap(_.getBlock(coords.getBlockRelColumn))
  def setBlock(block: Block): Boolean = getChunk(block.coord.getChunkRelWorld).fold(false)(_.setBlock(block))
  def removeBlock(coords: BlockRelWorld): Boolean = getChunk(coords.getChunkRelWorld).fold(false)(_.removeBlock(coords.getBlockRelChunk))

  def tick(camera: Camera): Unit = {
    chunkLoadingOrigin = CylCoord(camera.position.x, camera.position.y, camera.position.z)

    if (loadColumnsCountdown == 0) {
      loadColumnsCountdown = World.ticksBetweenColumnLoading

      chunksToLoad.enqueue(chunksToLoad.dequeueAll.map(c => makeChunkToLoadTuple(c._2)).filter(c => getColumn(c._2.getColumnRelWorld).isDefined): _*)

      updateLoadedColumns(camera)

      columns.values.foreach(_.updateLoadedChunks(camera))
    }
    loadColumnsCountdown -= 1

    if (!chunksToLoad.isEmpty) {
      val chunkToLoad = chunksToLoad.dequeue()._2
      getColumn(chunkToLoad.getColumnRelWorld).foreach(col => {
        val ch = chunkToLoad.getChunkRelColumn
        val chY = ch.Y
        col.topAndBottomChunks.foreach(topBottom => if (topBottom._1 >= chY && topBottom._2 <= chY) col.chunks(ch.value) = new Chunk(chunkToLoad, this))
      })
    }

    columns.values.foreach(_.tick())
  }

  private def updateLoadedColumns(camera: Camera): Unit = {
    val rDistSq = math.pow(renderDistance, 2)
    val origin = {
      val temp = chunkLoadingOrigin.toBlockCoord()
      new Vector2d(temp.x / 16, temp.z / 16)
    }
    def inSight(col: ColumnRelWorld): Boolean = {
      col.distSq(origin) <= rDistSq
    }

    val here = ColumnRelWorld(math.floor(origin.x).toInt, math.floor(origin.y).toInt)
    if (!columns.contains(here.value)) {
      columns(here.value) = new ChunkColumn(here, this)
      columnsAtEdge += here
    }
    val columnsToRemove = MutableSet.empty[ColumnRelWorld]
    val columnsToAdd = MutableSet.empty[ColumnRelWorld]
    for (col <- columnsAtEdge) {
      if (!inSight(col)) {
        columns.remove(col.value).get.unload()
        columnsToRemove += col
        for (offset <- ChunkColumn.neighbors) {
          val col2 = ColumnRelWorld(col.X + offset._1, col.Z + offset._2)
          if (columns.contains(col2.value)) {
            columnsToAdd += col2
          }
        }
      } else {
        var surrounded = true
        for (offset <- ChunkColumn.neighbors) {
          val col2 = ColumnRelWorld(col.X + offset._1, col.Z + offset._2)
          if (inSight(col2)) {
            if (!columns.contains(col2.value)) {
              columnsToAdd += col2
              columns(col2.value) = new ChunkColumn(col2, this)
            }
          } else {
            surrounded = false
          }
        }
        if (surrounded) columnsToRemove += col
      }
    }
    columnsAtEdge ++= columnsToAdd
    columnsAtEdge --= columnsToRemove
  }

  def unload(): Unit = {
    chunksToLoad.clear
    loadColumnsCountdown = -1
    columns.values.foreach(_.unload)
    columns.clear
  }
}
