package hexagon

import org.joml.Vector2f
import org.lwjgl.glfw.GLFW
import org.lwjgl.glfw.GLFW.GLFW_CURSOR
import org.lwjgl.glfw.GLFW.GLFW_CURSOR_DISABLED
import org.lwjgl.glfw.GLFW.GLFW_CURSOR_NORMAL
import org.lwjgl.glfw.GLFW.GLFW_KEY_ENTER
import org.lwjgl.glfw.GLFW.GLFW_KEY_ESCAPE
import org.lwjgl.glfw.GLFW.GLFW_KEY_M
import org.lwjgl.glfw.GLFW.GLFW_PRESS
import org.lwjgl.glfw.GLFW.glfwSetInputMode
import org.lwjgl.opengl.GL11

import hexagon.renderer.NoDepthTest
import hexagon.renderer.Renderer
import hexagon.renderer.VAOBuilder
import hexagon.renderer.VBO
import hexagon.resource.Shader
import hexagon.world.blockType.BlockType
import hexagon.world.coord.RayTracer
import hexagon.world.render.WorldRenderer
import hexagon.world.storage.Block
import hexagon.world.storage.World

trait Scene {
  /** Returns true to consume the event */
  def processKeys(key: Int, scancode: Int, action: Int, mods: Int): Boolean
  def processMouseButtons(button: Int, action: Int, mods: Int): Boolean
  def windowResized(w: Int, h: Int): Unit = {}
  def render(): Unit
  def tick(): Unit
  def unload(): Unit
  def onReloadedResources(): Unit
}

abstract class MenuScene extends Scene {
  val guiShader = Shader.getShader("gui")

  def unload(): Unit = {

  }
}

class MainMenuScene extends MenuScene {
  def processKeys(key: Int, scancode: Int, action: Int, mods: Int): Boolean = {
    if (action == GLFW_PRESS) key match {
      case GLFW_KEY_ENTER => Main.pushScene(new GameScene)
      case _              =>
    }
    true
  }

  def processMouseButtons(button: Int, action: Int, mods: Int): Boolean = {
    false
  }

  override def render(): Unit = {

  }

  override def tick(): Unit = {
  }
  
  def onReloadedResources(): Unit = ()
}

class PauseMenuScene extends MenuScene {
  def processKeys(key: Int, scancode: Int, action: Int, mods: Int): Boolean = {
    if (action == GLFW_PRESS) key match {
      case GLFW_KEY_ESCAPE => Main.popScene
      case _               =>
    }
    true
  }

  def processMouseButtons(button: Int, action: Int, mods: Int): Boolean = {
    false
  }

  override def render(): Unit = {

  }

  override def tick(): Unit = {
  }
  
  def onReloadedResources(): Unit = ()
}

class GameScene extends Scene {
  // Camera, player, mousepicker, world, etc.

  val blockShader = Shader.getShader("block").get
  val blockSideShader = Shader.getShader("blockSide").get
  val selectedBlockShader = Shader.getShader("selectedBlock").get
  val skyShader = Shader.getShader("sky").get
  val crosshairShader = Shader.getShader("crosshair").get

  private val crosshairVAO = new VAOBuilder(8).addVBO(VBO(4).floats(0, 2).create().fillFloats(0, Seq(0, 0.02f, 0, -0.02f, -0.02f, 0, 0.02f, 0))).create()
  private val crosshairRenderer = new Renderer(crosshairVAO, GL11.GL_LINES) with NoDepthTest

  val world = new World
  val worldRenderer = new WorldRenderer(world)

  val camera = new Camera(70f, Main.windowSize.x.toFloat / Main.windowSize.y, 0.02f, 1000f, world)
  val mousePicker = new RayTracer(world, camera, 7)
  val player = new Player(camera)

  private var leftMouseButtonDown = false
  private var rightMouseButtonDown = false
  private var leftMouseButtonCountdown = 0
  private var rightMouseButtonCountdown = 0

  setUniforms
  
  def onReloadedResources(): Unit = {
    setUniforms()
  }

  private def setUniforms(): Unit = {
    camera.setProjMatrix(blockShader)
    camera.setProjMatrix(blockSideShader)
    camera.setProjMatrix(selectedBlockShader)
    blockShader.setUniform1f("totalSize", World.totalSize)
    blockSideShader.setUniform1f("totalSize", World.totalSize)
    selectedBlockShader.setUniform1f("totalSize", World.totalSize)
    skyShader.setUniformMat4("invProjMatr", camera.invProjMatr)
    skyShader.setUniform2f("windowSize", Main.windowSize.x, Main.windowSize.y)
    crosshairShader.setUniform2f("windowSize", Main.windowSize.x, Main.windowSize.y)
  }

  def processKeys(key: Int, scancode: Int, action: Int, mods: Int): Boolean = {
    /* if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS) Main.pushScene(new PauseMenuScene)
    else */ if (key == GLFW_KEY_M && action == GLFW_PRESS) {
      player.moveWithMouse = !player.moveWithMouse
      glfwSetInputMode(Main.window, GLFW_CURSOR, if (player.moveWithMouse) GLFW_CURSOR_DISABLED else GLFW_CURSOR_NORMAL)
      Main.updateMousePos()
    } else if (key == GLFW_KEY_ENTER && action == GLFW_PRESS) {
      Main.popScene()
    }
    true
  }

  def processMouseButtons(button: Int, action: Int, mods: Int): Boolean = {
    if (button == 0) {
      rightMouseButtonDown = action != GLFW.GLFW_RELEASE
    } else if (button == 1) {
      leftMouseButtonDown = action != GLFW.GLFW_RELEASE
    }
    true
  }

  override def windowResized(width: Int, height: Int): Unit = {
    camera.aspect = width.toFloat / height
    camera.updateProjMatrix
    camera.setProjMatrix(blockShader)
    camera.setProjMatrix(blockSideShader)
    camera.setProjMatrix(selectedBlockShader)

    skyShader.setUniformMat4("invProjMatr", camera.invProjMatr)
    skyShader.setUniform2f("windowSize", width, height)
    crosshairShader.setUniform2f("windowSize", width, height)
  }

  def render(): Unit = {
    // render world etc.
    worldRenderer.render(camera)

    if (player.moveWithMouse) {
      crosshairShader.enable()
      crosshairRenderer.render()
    }
  }

  def tick(): Unit = {
    player.tick()
    camera.updateCoords
    camera.updateViewMatrix
    camera.updateUniforms(blockShader)
    camera.updateUniforms(blockSideShader)
    camera.updateUniforms(selectedBlockShader)
    skyShader.setUniformMat4("invViewMatr", camera.invViewMatr)
    skyShader.setUniform3f("sun", 0, 1, -1)
    blockShader.setUniform3f("sun", 0, 1, -1)
    blockSideShader.setUniform3f("sun", 0, 1, -1)

    // MousePicker
    mousePicker.setRayFromScreen(if (!player.moveWithMouse) Main.normalizedMousePos else new Vector2f(0, 0))
    worldRenderer.setSelectedBlockAndSide(mousePicker.trace(c => world.getBlock(c).isDefined))

    if (leftMouseButtonCountdown == 0) {
      if (leftMouseButtonDown) {
        leftMouseButtonCountdown = 10
        worldRenderer.getSelectedBlockAndSide match {
          case Some((coords1, Some(side))) =>
            val offset = Block.neighborOffsets(side)
            val coords = coords1.offset(offset._1, offset._2, offset._3)
            if (world.getBlock(coords).isEmpty) world.setBlock(new Block(coords, BlockType.Grass))
          case _ =>
        }
      }
    } else {
      leftMouseButtonCountdown -= 1
    }

    if (rightMouseButtonCountdown == 0) {
      if (rightMouseButtonDown) {
        rightMouseButtonCountdown = 10
        worldRenderer.getSelectedBlockAndSide match {
          case Some((coords, side)) =>
            if (world.getBlock(coords).isDefined) world.removeBlock(coords)
          case None =>
        }
      }
    } else {
      rightMouseButtonCountdown -= 1
    }
    world.tick(player)
  }

  def unload(): Unit = {
    glfwSetInputMode(Main.window, GLFW_CURSOR, GLFW_CURSOR_NORMAL)
  }
}
