package hexagon.world.storage

import java.util.concurrent.Executors

import scala.collection.mutable.PriorityQueue
import scala.collection.mutable.{ Set => MutableSet }
import scala.concurrent.ExecutionContext

import org.joml.Vector2d
import org.joml.Vector3d

import hexagon.Camera
import hexagon.world.coord.BlockRelWorld
import hexagon.world.coord.ChunkRelWorld
import hexagon.world.coord.ColumnRelWorld
import hexagon.world.coord.CoordUtils
import hexagon.world.coord.CylCoord
import hexagon.world.coord.BlockCoord
import java.util.Random
import hexagon.world.gen.noise.NoiseGenerator4D
import hexagon.world.gen.noise.NoiseGenerator3D
import hexagon.Player
import org.joml.Quaterniond
import org.joml.Vector4d

object World {
  /** Max-value: 20 */
  val worldSize = 7
  val ringSize = 1 << worldSize
  val ringSizeMask = ringSize - 1
  val totalSize = 16 * ringSize
  val totalSizeMask = totalSize - 1
  val hexAngle = (2 * math.Pi) / totalSize
  val radius = CoordUtils.y60 / hexAngle
  val circumference = totalSize * CoordUtils.y60
  val chunksLoadedPerTick = 2
  val chunkRenderUpdatesPerTick = 1
  val ticksBetweenColumnLoading = 5
}

class World {
  /* blockStorage
   * other world contents
   * methods for world interaction
   * method for loading/unloading blockColumns
   * block methods that will be forwarded to the correct chunk
   * worldGen references and methods
   * might rename this class to WorldStorage, and make another class World
   */

  var renderDistance: Double = 8 * CoordUtils.y60

  val columns = scala.collection.mutable.Map.empty[Long, ChunkColumn]
  val columnsAtEdge = MutableSet.empty[ColumnRelWorld]
  
  private val random = new Random(314151123)
  private[storage] val blockGenerator = new NoiseGenerator4D(random, 8, 0.1)
  private[storage] val heightMapGenerator = new NoiseGenerator3D(random, 8, 0.02)
  private[storage] val blockDensityGenerator = new NoiseGenerator4D(random, 4, 0.01)
  private[storage] val biomeHeightMapGenerator = new NoiseGenerator3D(random, 4, 0.002)
  private[storage] val biomeHeightVariationMapGenerator = new NoiseGenerator3D(random, 4, 0.002)

  private var loadColumnsCountdown = 0

  private[storage] var chunkLoadingOrigin: CylCoord = null
  private[storage] val chunkLoadingDirection: Vector3d = new Vector3d()
  private val chunksToLoad = PriorityQueue.empty[(Double, ChunkRelWorld)](Ordering.by(-_._1))
  val chunkLoader = ExecutionContext.fromExecutor(Executors.newFixedThreadPool(10))
  private val chunkRenderUpdateQueue = PriorityQueue.empty[(Double, ChunkRelWorld)](Ordering.by(-_._1))
  
  def addRenderUpdate(coords: ChunkRelWorld): Unit = {
    chunkRenderUpdateQueue.enqueue(makeChunkToLoadTuple(coords))
  }

  def addChunkToLoadingQueue(coords: ChunkRelWorld): Unit = {
    chunksToLoad.enqueue(makeChunkToLoadTuple(coords))
  }

  private def makeChunkToLoadTuple(coords: ChunkRelWorld) = {
    val cyl = BlockCoord(coords.X * 16 + 8, coords.Y * 16 + 8, coords.Z * 16 + 8).toCylCoord()
    val cDir = cyl.toNormalCoord(chunkLoadingOrigin).toVector3d.normalize()
    val dot = chunkLoadingDirection.dot(cDir)
    
    (chunkLoadingOrigin.distanceSq(cyl) * (1.25 - math.pow((dot + 1) / 2, 4)) / 1.25, coords)
  }

  def getColumn(coords: ColumnRelWorld): Option[ChunkColumn] = columns.get(coords.value)
  def getChunk(coords: ChunkRelWorld): Option[Chunk] = getColumn(coords.getColumnRelWorld).flatMap(_.getChunk(coords.getChunkRelColumn))
  def getBlock(coords: BlockRelWorld): Option[Block] = getColumn(coords.getColumnRelWorld).flatMap(_.getBlock(coords.getBlockRelColumn))
  def setBlock(block: Block): Boolean = getChunk(block.coord.getChunkRelWorld).fold(false)(_.setBlock(block))
  def removeBlock(coords: BlockRelWorld): Boolean = getChunk(coords.getChunkRelWorld).fold(false)(_.removeBlock(coords.getBlockRelChunk))

  def tick(player: Player): Unit = {
    chunkLoadingOrigin = CylCoord(player.camera.position.x, player.camera.position.y, player.camera.position.z)
    val vec4 = new Vector4d(0, 0, -1, 0).mul(player.camera.invViewMatr)
    chunkLoadingDirection.set(vec4.x, vec4.y, vec4.z)

    if (loadColumnsCountdown == 0) {
      loadColumnsCountdown = World.ticksBetweenColumnLoading
      
      val rDistSq = (renderDistance * 16) * (renderDistance * 16)

//      val startTime = System.nanoTime
//      chunksToLoad.enqueue(chunksToLoad.dequeueAll.map(c => makeChunkToLoadTuple(c._2)).filter(c => c._1 <= rDistSq): _*)
      chunksToLoad.clear()
//      println((System.nanoTime - startTime) * 1e-6)
//      println(chunksToLoad.size)
      chunkRenderUpdateQueue.enqueue(chunkRenderUpdateQueue.dequeueAll.map(c => makeChunkToLoadTuple(c._2)).filter(c => c._1 <= rDistSq): _*)

      updateLoadedColumns()

      columns.values.foreach(_.updateLoadedChunks())
    }
    loadColumnsCountdown -= 1
    //println(chunksToLoad.size + "\t" + chunkRenderUpdateQueue.size)
    for (_ <- 1 to World.chunksLoadedPerTick) {
      if (!chunksToLoad.isEmpty) {
        val chunkToLoad = chunksToLoad.dequeue()._2
        getColumn(chunkToLoad.getColumnRelWorld).foreach(col => {
          val ch = chunkToLoad.getChunkRelColumn
          val chY = ch.Y
          val topBottomChange = col.topAndBottomChunks match {
            case None => Some((chY, chY))
            case Some((top, bottom)) =>
              if (chY == top + 1) Some((chY, bottom))
              else if (chY == bottom - 1) Some((top, chY))
              else None
          }
          if (topBottomChange.isDefined) {
            col.topAndBottomChunks = topBottomChange
            col.chunks(ch.value) = new Chunk(chunkToLoad, this)
          }
        })
      }
    }
    
    for (_ <- 1 to World.chunkRenderUpdatesPerTick) {
      if (!chunkRenderUpdateQueue.isEmpty) {
        val chunk = chunkRenderUpdateQueue.dequeue()._2
        getChunk(chunk).foreach(_.doRenderUpdate())
      }
    }

    columns.values.foreach(_.tick())
  }

  private def updateLoadedColumns(): Unit = {
    val rDistSq = math.pow(renderDistance, 2)
    val origin = {
      val temp = chunkLoadingOrigin.toBlockCoord()
      new Vector2d(temp.x / 16, temp.z / 16)
    }
    def inSight(col: ColumnRelWorld): Boolean = {
      col.distSq(origin) <= rDistSq
    }

    val here = ColumnRelWorld(math.floor(origin.x).toInt, math.floor(origin.y).toInt)
    if (!columns.contains(here.value)) {
      columns(here.value) = new ChunkColumn(here, this)
      columnsAtEdge += here
    }
    val columnsToRemove = MutableSet.empty[ColumnRelWorld]
    val columnsToAdd = MutableSet.empty[ColumnRelWorld]
    for (col <- columnsAtEdge) {
      if (!inSight(col)) {
        columns.remove(col.value).get.unload()
        columnsToRemove += col
        for (offset <- ChunkColumn.neighbors) {
          val col2 = ColumnRelWorld(col.X + offset._1, col.Z + offset._2)
          if (columns.contains(col2.value)) {
            columnsToAdd += col2
          }
        }
      } else {
        var surrounded = true
        for (offset <- ChunkColumn.neighbors) {
          val col2 = ColumnRelWorld(col.X + offset._1, col.Z + offset._2)
          if (inSight(col2)) {
            if (!columns.contains(col2.value)) {
              columnsToAdd += col2
              columns(col2.value) = new ChunkColumn(col2, this)
            }
          } else {
            surrounded = false
          }
        }
        if (surrounded) columnsToRemove += col
      }
    }
    columnsAtEdge ++= columnsToAdd
    columnsAtEdge --= columnsToRemove
  }

  def unload(): Unit = {
    chunksToLoad.clear
    loadColumnsCountdown = -1
    columns.values.foreach(_.unload)
    columns.clear
  }
}
